
-----------------------------------------------------------------------------

local mseq = 0 -- message sequence number
local version, build, date, tocversion = GetBuildInfo()

-----------------------------------------------------------------------------

local L = DisplacedEnergyWatcherLocale.DisplacedEnergyWatcher
if not L then return end

-- global on purpose
DisplacedEnergyWatcher = LibStub("AceAddon-3.0"):NewAddon("DisplacedEnergyWatcher", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0", "LibSink-2.0")

local isConfig = false
local AceConfig = LibStub("AceConfig-3.0")

local isDialog = false
local AceDialog = LibStub("AceConfigDialog-3.0")

local isDBOpt = false
local AceDBOptions = LibStub("AceDBOptions-3.0")

local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")

-----------------------------------------------------------------------------

DisplacedEnergyWatcher.namespace = "displacedEnergyWatcher"
DisplacedEnergyWatcher.defaults =
{
	profile =
	{
		enable		= true,

		warnAfflict = false,
		warnDamage	= true,
		warnDispel	= true,

		whisper		= false,
		chatFrame	= "ChatFrame1",
	},
}

-----------------------------------------------------------------------------

local N = nil
local DisplacedEnergyWatcher = DisplacedEnergyWatcher

-- Blizzard API upvalues rather than a global lookup
local SendChatMessage = SendChatMessage
local CheckInteractDistance = CheckInteractDistance
local UnitGUID = UnitGUID
local UnitName = UnitName
local time = time

-- Since pairs and select are used often and in loops, make them upvalues rather than a global lookup
local pairs = pairs
local select = select

-----------------------------------------------------------------------------

-- simple list of unitids
local Units = { }

-- reverse lookups from GUID to unitid, for CheckInteractDistance use
local UnitGUIdRMap = { }

-- lists for tracking clean, afflicted, and removed GUIDs
local UnitCleanList = { }
local UnitAfflictedList = { }
local UnitRemovedList = { }

local ChatFrames =
{
	["ChatFrame1"] = ChatFrame1,
	["ChatFrame2"] = ChatFrame2,
	["ChatFrame3"] = ChatFrame3,
	["ChatFrame4"] = ChatFrame4,
	["ChatFrame5"] = ChatFrame5,
	["ChatFrame6"] = ChatFrame6,
	["ChatFrame7"] = ChatFrame7,
}

-----------------------------------------------------------------------------

local PLAYER_GUID = nil

local MAX_IN_PARTY = 4
local MAX_IN_RAID = 40

local SPELL_ID_DISPLACED_ENERGY = 142913 -- Displaced Energy spell ID; used for both debuff/aura and periodic damage

local DAMAGE_RANGE = 2 -- CheckInteractDistance using "trade range", documented at 11.11 yards

local STATE_AFFLICT = 0	-- initial affliction
local STATE_DAMAGE3 = 3	-- at 3 seconds
local STATE_DAMAGE6 = 6	-- at 6 seconds
local STATE_DAMAGE9 = 9	-- at 9 seconds

local MESSAGE_BOILER = "DislacedEnergyWatcher: |cffff0000%s|r" -- warning message text color hardcoded to red, with default text color preamble

-----------------------------------------------------------------------------
-- generate warning messages to the appropriate message sink(s)
local function BroadcastMessage(message)
	if message then
		DisplacedEnergyWatcher:Pour(message, 1, 0, 0) -- LibSink-2.0

		-- chat frame duplicate; only if LibSink-2.0 is not already sending to default chat frame
		local profile = DisplacedEnergyWatcher.db.profile
		local chatEnabled = profile.sink20OutputSink ~= "ChatFrame" and profile.chatFrame ~= "None"
		if chatEnabled then
			local chatFrame = ChatFrames[DisplacedEnergyWatcher.db.profile.chatFrame]
			if chatFrame then
				chatFrame:AddMessage(message, 1, 0, 0)
			end
		end
	end
end

-- whisper a message to an offending player
local function WhisperMessage(unitName, message)
	if DisplacedEnergyWatcher.db.profile.whisper then
		if unitName and message then
			SendChatMessage(message, "WHISPER", nil, unitName)
		end
	end
end

-----------------------------------------------------------------------------

function DisplacedEnergyWatcher:OnInitialize()
	-- mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:OnInitialize(): enter")

	if not N then N = L["DisplacedEnergyWatcher"] .. "-" .. DisplacedEnergyWatcher.version end

	DisplacedEnergyWatcher.DisplacedEnergyWatcherDB = LibStub("AceDB-3.0"):New("DisplacedEnergyWatcherDB", nil, "Default")
	if not self.db or self.db ~= DisplacedEnergyWatcher.DisplacedEnergyWatcherDB:GetNamespace(self.namespace, true) then self.db = DisplacedEnergyWatcher.DisplacedEnergyWatcherDB:RegisterNamespace(self.namespace, self.defaults) end

	-- mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:OnInitialize(): exit")
end

function DisplacedEnergyWatcher:OnEnable()
	-- mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:OnEnable(): enter")

	-- configure options
	if (not isConfig) then
		AceConfig:RegisterOptionsTable(N, DisplacedEnergyWatcher.Options, "displacedEnergyWatcher")
		-- ChatFrame7:AddMessage("DisplacedEnergyWatcher:OnEnable(): register options " .. "'" .. tostring(N) .. "'")
		isConfig = true
	end

	if (not isDialog) then
		self.optionsFrame = AceDialog:AddToBlizOptions(N, "DisplacedEnergyWatcher")
		isDialog = true
	end

	if (not isDBOpt) then
		DisplacedEnergyWatcher.Options.args.profiles = AceDBOptions:GetOptionsTable(DisplacedEnergyWatcher.DisplacedEnergyWatcherDB)
		DisplacedEnergyWatcher.Options.args.profiles.order = 800
		isDBOpt = true
	end

	-- configure LibSink-2.0 options; add a disabled function
	self.Options.args.output = self:GetSinkAce3OptionsDataTable()
	self.Options.args.output.disabled = function(i) return not self.db.profile.enable end

	self:SetSinkStorage(self.db.profile)
	-- TODO: add an option to whisper offending players

	PLAYER_GUID = UnitGUID("player")
	self:Connect()
	-- mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:OnEnable(): exit")
end

function DisplacedEnergyWatcher:OnDisable()
	-- mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:OnDisable(): enter")

	self:Disconnect()

	-- mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:OnDisable(): exit")
end

function DisplacedEnergyWatcher:Update()
	mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:Update(): enter")

	if self.db.profile.enable then
		self:Connect()
	else
		self:Disconnect()
	end

	mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:Update(): exit")
end

function DisplacedEnergyWatcher:Connect()
	-- register event
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "OnCombatLogEvent") -- combat log event
	self:RegisterEvent("GROUP_ROSTER_UPDATE", "OnGroupRosterUpdate") -- player added/moved/removed from group

	self:RegisterEvent("PLAYER_REGEN_DISABLED", "OnPlayerRegenDisabled") -- combat starts
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "OnPlayerRegenEnabled") -- combat ends

	-- configure initial unit GUID -> unitid map
	self:ScanFullRoster()
end

function DisplacedEnergyWatcher:Disconnect()
	self:UnregisterAllEvents()
	self:UnregisterAllMessages()
	self:CancelAllTimers()
end

-----------------------------------------------------------------------------

-- TODO: make extensible by splitting all of this out into a Malkorok module?

local function spell_dispel(timestamp, subtype, originGUId, originName, originFlags, originFlags2, targetGUId, targetName, targetFlags, targetFlags2, spellId, spellName, spellSchool)
	local ts = time()

	if targetGUId ~= PLAYER_GUID and UnitRemovedList[targetGUId] then
		local unitid = UnitGUIdRMap[targetGUId]

		mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:SPELL_DISPEL(): " .. tostring(originName) .. "/" .. tostring(originGUId)
			.. " " .. tostring(spellName) .. "/" .. tostring(spellId)
			.. " dispelled " .. "/" .. tostring(targetName) .. "/" .. tostring(targetGUId) .. " @ " .. tostring(unitid) .. "/" .. tostring(unitid and CheckInteractDistance(unitid, DAMAGE_RANGE)))

		if DisplacedEnergyWatcher.db.profile.warnDispel and unitid and CheckInteractDistance(unitid, DAMAGE_RANGE) then
			local message = L["STOP IT"] .. ": " .. originName .. " " .. L["DISPELLED Displaced Energy TOO CLOSE TO OTHERS!!!"]
			mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:SPELL_DISPEL(): " .. MESSAGE_BOILER:format(message))

			WhisperMessage(originName, message)
			BroadcastMessage(message)
		end
	end

	UnitRemovedList[targetGUId] = nil
end

local function spell_periodic_damage(timestamp, subtype, originGUId, originName, originFlags, originFlags2, targetGUId, targetName, targetFlags, targetFlags2, spellId, spellName, spellSchool, spellDamage, spellOverkill, school, resisted, blocked, absorbed, critical, glancing, crushing)
	-- Malkorok has done damage to SOMEONE, spit out a diagnostic
	if spellId == SPELL_ID_DISPLACED_ENERGY and originName == L["Malkorok"] then
		local unitid = UnitGUIdRMap[targetGUId] -- get the afflicted unit's unitid
		-- mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:SPELL_PERIODIC_DAMAGE(): " .. tostring(subtype) .. "/" .. tostring(spellId) .. "/" .. tostring(spellName) .. "/" .. tostring(spellSchool) .. " || " .. tostring(originGUId) .. "/" .. tostring(originName) .. " -> " .. tostring(targetGUId) .. "/" .. tostring(targetName))
		mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:SPELL_PERIODIC_DAMAGE(): " .. tostring(originName) .. "/" .. tostring(originGUId)
			.. " " .. tostring(spellName) .. "/" .. tostring(spellId) .. " damages " .. "/" .. tostring(targetName) .. "/" .. tostring(targetGUId) .. " @ " .. tostring(unitid) .. "/" .. tostring(unitid and CheckInteractDistance(unitid, DAMAGE_RANGE)))
	end

	-- if "Displaced Energy" (by spellid) from "Malkorok" does damage to ME...
	if DisplacedEnergyWatcher.db.profile.warnDamage and spellId == SPELL_ID_DISPLACED_ENERGY and originName == L["Malkorok"] and targetGUId == PLAYER_GUID then
		-- iterate all units currently afflicted, to see if any are close enough to ME to cause damage, and append those names together
		local names = nil
		for unitGUId in pairs(UnitAfflictedList) do
			local unitid = UnitGUIdRMap[unitGUId] -- get the afflicted unit's unitid
			if unitid and CheckInteractDistance(unitid, DAMAGE_RANGE) then
				local n, r = UnitName(unitid)

				-- getting a "attempt to concatenate nil value" here
				if n then
					local fn = nil
					if r then
						fn = n .. "-" .. "r"
					else
						fn = n
					end

					if fn then
						-- whisper to individual player
						WhisperMessage(fn, L["GET AWAY"] .. ": " .. fn .. " " .. L["Displaced Energy is DAMAGING OTHERS!!!"])

						-- accumulate all potential culprits, for global message
						if names then
							names = names .. ", " .. fn
						else
							names = fn
						end
					end
				end
			end
		end

		-- if we found some unit names, generate a message
		if names then
			-- if an afflicted unit is close enough to me
			local message = L["GET AWAY"] .. ": " .. names .. " " .. L["Displaced Energy is DAMAGING OTHERS!!!"]
			mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:SPELL_AURA_APPLIED(): " .. MESSAGE_BOILER:format(message))

			BroadcastMessage(message)
		end
	end
end

local function spell_aura_applied(timestamp, subtype, originGUId, originName, originFlags, originFlags2, targetGUId, targetName, targetFlags, targetFlags2, spellId, spellName, spellSchool, auraType)
	-- if originName == L["Malkorok"] then
	--	mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:SPELL_AURA_APPLIED(): " .. tostring(subtype) .. "/" .. tostring(spellId) .. "/" .. tostring(spellName) .. "/" .. tostring(auraType) .. " || " .. tostring(originGUId) .. "/" .. tostring(originName) .. " -> " .. tostring(targetGUId) .. "/" .. tostring(targetName))
	-- end

	if spellId == SPELL_ID_DISPLACED_ENERGY and originName == L["Malkorok"] then
		mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:SPELL_AURA_APPLIED(): " .. tostring(originName) .. "/" .. tostring(originGUId)
			.. " " .. tostring(spellName) .. "/" .. tostring(spellId)
			.. " afflicts " .. "/" .. tostring(targetName) .. "/" .. tostring(targetGUId))

		if DisplacedEnergyWatcher.db.profile.warnAfflict then
			UnitCleanList[targetGUId] = nil
			UnitAfflictedList[targetGUId] = true
			UnitRemovedList[targetGUId] = nil

			local message = L["GET AWAY"] .. ": " .. targetName .. " " .. L["AFFLICTED with Displaced Energy!!!"]
			mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:SPELL_AURA_APPLIED(): " .. MESSAGE_BOILER:format(message))

			WhisperMessage(targetName, message)
			BroadcastMessage(message)
		end
	end
end

local function spell_aura_removed(timestamp, subtype, originGUId, originName, originFlags, originFlags2, targetGUId, targetName, targetFlags, targetFlags2, spellId, spellName, spellSchool, auraType)
	-- if originName == L["Malkorok"] then
	--	mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:SPELL_AURA_REMOVED(): " .. tostring(subtype) .. "/" .. tostring(spellId) .. "/" .. tostring(spellName) .. "/" .. tostring(auraType) .. " || " .. tostring(originGUId) .. "/" .. tostring(originName) .. " -> " .. tostring(targetGUId) .. "/" .. tostring(targetName))
	-- end

	if spellId == SPELL_ID_DISPLACED_ENERGY and originName == L["Malkorok"] then
		mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:SPELL_AURA_REMOVED(): " .. tostring(originName) .. "/" .. tostring(originGUId)
			.. " " .. tostring(spellName) .. "/" .. tostring(spellId)
			.. " fades from " .. "/" .. tostring(targetName) .. "/" .. tostring(targetGUId))

		UnitCleanList[targetGUId] = true
		UnitAfflictedList[targetGUId] = nil
		UnitRemovedList[targetGUId] = time() -- current time, in sections; TODO: could probably be "true"
	end
end

local function unit_died(timestamp, subtype, originGUId, originName, originFlags, originFlags2, targetGUId, targetName, targetFlags)
	-- mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:UNIT_DIED(): " .. tostring(subtype) .. " || " .. tostring(originGUId) .. "/" .. tostring(originName) .. " -> " .. tostring(targetGUId) .. "/" .. tostring(targetName))

	local unitid = UnitGUIdRMap[targetGUId] -- TODO: is this even necessary?
	if unitid then
		UnitCleanList[targetGUId] = true
		UnitAfflictedList[targetGUId] = nil
		UnitRemovedList[targetGUId] = nil
	end
end

-----------------------------------------------------------------------------

local subtypeDispatchMethods =
{
	-- prefix: SPELL_
	["SPELL_DISPEL"]			= spell_dispel,

	-- prefix: SPELL_PERIODIC_
	["SPELL_PERIODIC_DAMAGE"]	= spell_periodic_damage,

	-- prefix: SPELL_AURA_
	["SPELL_AURA_APPLIED"]		= spell_aura_applied,
	["SPELL_AURA_REMOVED"]		= spell_aura_removed,

	-- unit death
	["UNIT_DIED"]				= unit_died,
}

-----------------------------------------------------------------------------

function DisplacedEnergyWatcher:OnCombatLogEvent(event, timestamp, subtype, hideCaster, originGUId, originName, originFlags, originFlags2, targetGUId, targetName, targetFlags, targetFlags2, suffix1, suffix2, suffix3, suffix4, suffix5, suffix6, suffix7, suffix8, suffix9, suffix10, suffix11, suffix12)
	--[[
		mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:OnCombatLogEvent(): " .. tostring(timestamp) .. "/" .. tostring(subtype) .. "|"
			.. tostring(originGUId) .. "/" .. tostring(originName) .. "/" .. tostring(originFlags) .. "|"
			.. tostring(targetGUId) .. "/" .. tostring(targetName) .. "/" .. tostring(targetFlags) .. "|"
			.. tostring(spellId) .. "/" .. tostring(spellName) .. "|"
			.. tostring(suffix1) .. "/" .. tostring(suffix2) .. "/" .. tostring(suffix3) .. "/" .. tostring(suffix4) .. "/" .. tostring(suffix5) .. "/" .. tostring(suffix6) .. "/" .. tostring(suffix7) .. "/" .. tostring(suffix8) .. "/" .. tostring(suffix9) .. "/" .. tostring(suffix10) .. "/" .. tostring(suffix11) .. "/" .. tostring(suffix12))
	end
	]]

	local dispatchMethod = subtypeDispatchMethods[subtype]
	if dispatchMethod then
		dispatchMethod(timestamp, subtype, originGUId, originName, originFlags, originFlags2, targetGUId, targetName, targetFlags, targetFlags2, suffix1, suffix2, suffix3, suffix4, suffix5, suffix6, suffix7, suffix8, suffix9, suffix10, suffix11, suffix12)
	end
end

-----------------------------------------------------------------------------

function DisplacedEnergyWatcher:OnGroupRosterUpdate(event)
	self:ScheduleTimer(function() DisplacedEnergyWatcher:ScanFullRoster() end, 0.2)
end

-- initialize roster reverse mapping, unitGUID->unitid, using the most generic available unitid
function DisplacedEnergyWatcher:ScanFullRoster()
	-- zap existing maps
	Units = { }
	UnitGUIdRMap = { }

	-- map all raid unitids; if active, these will be preferred
	for i = 1, 40 do
		local unitid = "raid" .. i
		self:RMapUnitId(unitid)
	end

	-- map party unitids, if necessary
	for i = 1, 4 do
		local unitid = "party" .. i
		self:RMapUnitId(unitid)
	end

	-- map solo player unitid, if necessary
	self:RMapUnitId("player")
end

-- reverse-map unitids from GUIDs; unitids are more compact, but GUIDs are used in the combat log
function DisplacedEnergyWatcher:RMapUnitId(unitid)
	if UnitExists(unitid) then
		local unitGUId = UnitGUID(unitid)
		if unitGUId and not UnitGUIdRMap[unitGUId] then
			Units[unitid] = true
			UnitGUIdRMap[unitGUId] = unitid
			-- mseq = mseq + 1; ChatFrame7:AddMessage(("%08d "):format(mseq) .. "DisplacedEnergyWatcher:RMapUnitId(): " .. tostring(unitid) .. "/" .. tostring(unitGUId))
		end
	end
end

-----------------------------------------------------------------------------

-- combat starts
function DisplacedEnergyWatcher:OnPlayerRegenDisabled(event)
	-- zap all of our tables
	UnitCleanList = { }
	UnitAfflictedList = { }
	UnitRemovedList = { }

	-- add all unit GUIDs to clean list
	for unitid in pairs(Units) do
		local unitGUId = UnitGUID(unitid)
		UnitCleanList[unitGUId] = true
	end
end

-- combat ends
function DisplacedEnergyWatcher:OnPlayerRegenEnabled(event)
	-- UnitCleanList = { }
	-- UnitAfflictedList = { }
end

-----------------------------------------------------------------------------